//
//  SquishController.mm
//  squish
//
//  Created by Da Woon Jung on 2006-06-23.
//  Copyright 2006-9 Da Woon Jung. All rights reserved.
//

#import <Carbon/Carbon.h>
#import <QuickTime/QuickTime.h>
#include <unistd.h>
#import "SquishController.h"
#import "SquishExportOptionsController.h"
#import "SquishProgressController.h"
#import "SquishCompressNotifier.h"
#import "SquishDDSWindow.h"
#import "GIWrapper.h"
#import "dds.h"
#import "imageutils.h"
#import <squish/squish.h>
using namespace squish;


#define DDS_EXT     @"dds"
#define DXT5NM_EXT  @"dxt5nm"
NSString *MIPMAP_EXT[] = {
    @"png",
    @"tiff"
};

enum {
    kDxt5nm = 1 << 3
};

enum {
    kFileModeDDS,
    kFileModeMipmap
};

@interface SquishController (Private)
- (NSApplicationDelegateReply)exportTilesConfirmingReplace: (NSArray *)aTiles
                                                    withGI: (GIWrapper *)aGi
                                                    toFile: (NSString *)aOutFile
                                                    ofType: (NSString *)aType
                                                 withFlags: (int)aFlags
                                                  fileMode: (int)aFileMode
                                               refreshTime: (NSDate **)aTime;
- (NSApplicationDelegateReply)exportTiles: (NSArray *)aTiles
                                   withGI: (GIWrapper *)aGi
                                   toFile: (NSString *)aOutFile
                                   ofType: (NSString *)aType
                                withFlags: (int)aFlags
                                 fileMode: (int)aFileMode;
- (NSArray *)extensions;
- (int)flagsFromPrefs;
- (NSString *)pathForOutputFile: (NSString *)aFile
                         suffix: (NSString *)aSuffix
                      extension: (NSString *)aExt
                   outputFolder: (NSString *)aDest;
- (NSApplicationDelegateReply)compressBytes: (unsigned char *)aBytes
                                    toBlock: (u8 *)output
                                      width: (size_t)aWidth
                                     height: (size_t)aHeight
                                bytesPerRow: (size_t)aBytesPerRow
                            samplesPerPixel: (int)aSamplesPerPixel
                                      alpha: (int)aAlpha
                                  withFlags: (int)aFlags;
- (NSApplicationDelegateReply)compressBytes: (unsigned char *)aBytes
                                     toFile: (FILE *)ofp
                                      width: (size_t)aWidth
                                     height: (size_t)aHeight
                                bytesPerRow: (size_t)aBytesPerRow
                            samplesPerPixel: (int)aSamplesPerPixel
                                      alpha: (int)aAlpha
                                premultiply:
#ifdef USE_AUTO_PREMULTIPLY
                                             (BOOL)aPremultiply
#else
                                             (int)aPremulOption
#endif
                                  withFlags: (int)aFlags;
- (NSApplicationDelegateReply)writeDDSHeaderToFile: (FILE *)ofp
                                            ofType: (NSString *)aType
                                             width: (size_t)aWidth
                                            height: (size_t)aHeight
                                         withFlags: (int)aFlags
                                           mipmaps: (BOOL)genMipmaps;
- (NSApplicationDelegateReply)decompressBytes: (u8 *)aBytes
                                      toBlock: (u8 **)output
                                        width: (size_t)aWidth
                                       height: (size_t)aHeight
                                    withFlags: (int)aFlags;
@end


@implementation SquishController

- (void)awakeFromNib
{
    sysprof = CreateSystemProfile();
    cspace = CGColorSpaceCreateWithPlatformColorSpace(sysprof);
}

- (void)applicationWillFinishLaunching:(NSNotification *)aNotification
{
    [SquishExportOptionsController registerDefaults];
    showOptions = (0 != (GetCurrentKeyModifiers() & optionKey));
}

- (void)applicationDidFinishLaunching:(NSNotification *)aNotification
{
    if (didOpenFiles)
        [[aNotification object] performSelector: @selector(terminate:) withObject: nil afterDelay: 0.1];
    else
        showOptions = YES;
}

- (void)applicationWillTerminate:(NSNotification *)aNotification
{
#ifdef USE_DDS_PREVIEW
    [ddsPreview release];
#endif
    [exportController release];
    CMCloseProfile(sysprof);
    CGColorSpaceRelease(cspace);
}

- (BOOL)applicationShouldOpenUntitledFile:(NSApplication *)aSender
{
    return NO;
}

- (void)application:(NSApplication *)aSender openFiles:(NSArray *)aFilenames
{
    NSApplicationDelegateReply reply = NSApplicationDelegateReplyFailure;
    didOpenFiles = YES;
    [self setEnableGenMipmaps: YES];
    [self setEnableAssembleMipmaps: YES];

    if (aFilenames && [aFilenames count] > 0)
    {
        if ([aFilenames count] < 2)
        {
            [[NSUserDefaults standardUserDefaults] setBool:NO forKey:@"assembleMipmaps"];
            [self setEnableAssembleMipmaps: NO];
        }

        if (showOptions)
        {
            if (NSRunAbortedResponse == [self showExportOptions])
            {
                // Do not reply "NSApplicationDelegateReplyCancel"
                // because doing so causes the filename to disappear
                // from the Recent Items list!
                [aSender replyToOpenOrPrint: NSApplicationDelegateReplySuccess];
                return;
            }
        }

        if (NSApplicationDelegateReplyCancel != reply)
        {
            reply = [self exportFiles: aFilenames];
        }
    }

    [aSender replyToOpenOrPrint: reply];
}


- (void)browseFilesOrFolders: (id)aSender
{
    NSOpenPanel *oPanel = [NSOpenPanel openPanel];
    [oPanel setAllowsMultipleSelection: YES];
    [oPanel setCanChooseDirectories: YES];
    [oPanel setCanChooseFiles: YES];
    
    if (NSFileHandlingPanelOKButton ==
        [oPanel runModalForDirectory: nil
                                file: nil
                               types: [self extensions]])
    {
        [self application: NSApp openFiles: [oPanel filenames]];
    }
}

- (int)showExportOptions
{
    if (nil == exportController)
        exportController =
            [[SquishExportOptionsController alloc] init];

    int result = [NSApp runModalForWindow: [exportController window]];
    [[exportController window] orderOut: self];
    return result;
}


- (NSApplicationDelegateReply)exportFiles: (NSArray *)aFilenames
{
    NSFileManager *fm = [NSFileManager defaultManager];
    NSUserDefaults *defaults = [NSUserDefaults standardUserDefaults];
    NSEnumerator *selectionEnum = [aFilenames objectEnumerator];
    NSDirectoryEnumerator *dirEnum;
    NSString *path;
    NSString *file;
    BOOL isDirectory;
    NSSet *extSet = [NSSet setWithArray: [self extensions]];
    int flags = [self flagsFromPrefs];
    int fileMode = [defaults integerForKey: @"fileMode"];
    int mipmapFormat = [defaults integerForKey: @"mipmapFormat"];
    BOOL assembleMipmaps = NO;
    NSString *outputType = nil;
    NSString *dest = [defaults stringForKey: @"browseDest"];
    NSString *destSuffix = @"";
    NSString *destExt = DDS_EXT;

    switch (fileMode)
    {
        case kFileModeDDS:
            switch ([defaults integerForKey: @"outputFormat"])
            {
            case kDxt3:
                outputType = @"DXT3"; break;
            case kDxt5:
                outputType = @"DXT5"; break;
            case kDxt5nm:
                outputType = @"DXT5"; destExt = DXT5NM_EXT; break;
            default:
                outputType = @"DXT1";
            }
            assembleMipmaps = [defaults boolForKey: @"assembleMipmaps"];
            break;
        case kFileModeMipmap:
            outputType = @"mipmaps";
            if (mipmapFormat < 0 || mipmapFormat > kMipmapTIFF)
                mipmapFormat = kMipmapPNG;
            destExt = MIPMAP_EXT[mipmapFormat];
            break;
        default:
            break;
    }
//    NSLog(@"Format = '%@'", outputType);
    NSMutableArray *folderContents = [NSMutableArray array];

    while ((path = [selectionEnum nextObject]))
    {
        if ([fm fileExistsAtPath: path isDirectory: &isDirectory])
        {
            if (isDirectory)
            {
                dirEnum = [fm enumeratorAtPath: path];
                while ((file = [dirEnum nextObject]))
                {
                    if ([extSet containsObject: [file pathExtension]])
                    {
                        [folderContents addObject: [path stringByAppendingPathComponent: file]];
                    }
                }
            }
            else
            {
                [folderContents addObject: path];
            }
        }
    }
#ifdef USE_DDS_PREVIEW
    if (nil == ddsPreview)
        ddsPreview = [[SquishDDSWindow alloc] init];
#endif

    int theReplaceChoice = kSquishNoReplace;
    progressController = [[SquishProgressController alloc] init];
    [progressController window];
    [progressController showWindow: self];
    [progressController setProgressIndeterminate: NO];
    [progressController setProgressItemCount: [folderContents count]];
    [progressController setProgressTitleText: [NSString stringWithFormat: NSLocalizedString(@"Exporting to %@...",""), outputType]];
    [progressController setProgressItemText: @""];
    [progressController startProgress];

    NSEnumerator *contentEnum;
    NSAutoreleasePool *pool;
    NSString *outFile;
    NSDate *curTime = [[NSDate alloc] init];
    NSApplicationDelegateReply exitStatus = NSApplicationDelegateReplySuccess;
//    Handle srcDataRef = NULL;
//    OSType srcDataRefType;
//    Rect srcRect;
//    size_t srcWidth = 0, srcHeight = 0;
    NSSize srcSize = NSZeroSize;
    BOOL mipmapsPossible = NO;
//    ComponentInstance gi = NULL;
//    OSErr err = noErr;
    GIWrapper *gi = nil;

    contentEnum = [folderContents objectEnumerator];

    if (assembleMipmaps)
    {
        NSMutableDictionary *tileDict = [NSMutableDictionary dictionaryWithCapacity: [folderContents count]];
        NSMutableArray *sortedPaths = nil;
        NSArray *sortedTileKeys = nil;
        NSString *tileKey;

        while ((path = [contentEnum nextObject]))
        {
            gi = [[GIWrapper alloc] initWithPath: path];
            if (nil == gi)
            {
                exitStatus = NSApplicationDelegateReplyFailure;
                break;
            }

            srcSize = [gi size];
            tileKey = [NSString stringWithFormat:@"%010.0f %010.0f", srcSize.width, srcSize.height];
            // Duplicates will be silently ignored
            [tileDict setObject:path forKey:tileKey];
            [gi release];
        }

        if (NSApplicationDelegateReplySuccess == exitStatus)
        {
            sortedPaths = [NSMutableArray arrayWithCapacity: [folderContents count]];
            sortedTileKeys = [[tileDict allKeys] sortedArrayUsingSelector:@selector(compare:)];
            NSEnumerator *tileEnum = [sortedTileKeys reverseObjectEnumerator];
            NSArray *dimensions;
            tileKey = [tileEnum nextObject];
            dimensions = [tileKey componentsSeparatedByString: @" "];
            srcSize.width  = [[dimensions objectAtIndex: 0] intValue];
            srcSize.height = [[dimensions objectAtIndex: 1] intValue];
            [sortedPaths addObject: [tileDict objectForKey: tileKey]];

            while ((tileKey = [tileEnum nextObject]))
            {
                if (srcSize.width  > 1) { srcSize.width  *= .5f; }
                if (srcSize.height > 1) { srcSize.height *= .5f; }
                dimensions = [tileKey componentsSeparatedByString: @" "];
                if (srcSize.width  != [[dimensions objectAtIndex: 0] intValue] ||
                    srcSize.height != [[dimensions objectAtIndex: 1] intValue])
                {
                    exitStatus = NSApplicationDelegateReplyFailure;
                    break;
                }
                [sortedPaths addObject: [tileDict objectForKey: tileKey]];
            }
            if (srcSize.width > 1 || srcSize.height > 1)
                exitStatus = NSApplicationDelegateReplyFailure;
            if (NSApplicationDelegateReplySuccess == exitStatus)
            {
                path = [sortedPaths objectAtIndex: 0];
                outFile = [self pathForOutputFile: path
                                           suffix: @""
                                        extension: destExt
                                     outputFolder: dest];
                [progressController setProgressItemText: [path lastPathComponent]];
                [progressController setProgressItemCount: 1];
                if ([progressController incrementProgress])
                    exitStatus = [self exportTilesConfirmingReplace:sortedPaths withGI:nil toFile:outFile ofType:outputType withFlags:flags fileMode:fileMode refreshTime:&curTime];
            }
        }
    }
    else
    {
        while (NSApplicationDelegateReplySuccess == exitStatus &&
               (path = [contentEnum nextObject]))
        {
            [progressController setProgressItemText: [path lastPathComponent]];
            if (![progressController incrementProgress])
                break;
            
            pool = [[NSAutoreleasePool alloc] init];
            mipmapsPossible = NO;
            destSuffix = @"";
            
            gi = [[[GIWrapper alloc] initWithPath: path] autorelease];
            if (nil == gi)
            {
                [pool release];
                exitStatus = NSApplicationDelegateReplyFailure;
                break;
            }
            
            if (kFileModeMipmap == fileMode)
            {
                srcSize = [gi size];
                mipmapsPossible = (srcSize.width > 0 && srcSize.height > 0) && ((int)srcSize.width % 2 == 0) && ((int)srcSize.height % 2 == 0);
                if (!mipmapsPossible)
                {
                    [pool release];
                    exitStatus = NSApplicationDelegateReplyFailure;
                    break;
                }
            }
            
            if (mipmapsPossible)
            {
                destSuffix = [NSString stringWithFormat: @"_%.0fx%.0f", srcSize.width, srcSize.height];
            }
            
            outFile = [self pathForOutputFile: path
                                       suffix: destSuffix
                                    extension: destExt
                                 outputFolder: dest];
            exitStatus = [self exportTilesConfirmingReplace:[NSArray arrayWithObject: path] withGI:gi toFile:outFile ofType:outputType withFlags:flags fileMode:fileMode refreshTime: &curTime];
            if (NSApplicationDelegateReplySuccess == exitStatus &&
                mipmapsPossible)
            {
                while (srcSize.width > 1 || srcSize.height > 1)
                {
                    if (srcSize.width  > 1) { srcSize.width  *= .5f; }
                    if (srcSize.height > 1) { srcSize.height *= .5f; }
                    destSuffix = [NSString stringWithFormat: @"_%.0fx%.0f", srcSize.width, srcSize.height];
                    outFile = [self pathForOutputFile: path
                                               suffix: destSuffix
                                            extension: destExt
                                         outputFolder: dest];
                    
                    [gi setSize: srcSize];
                    exitStatus = [self exportTilesConfirmingReplace:[NSArray arrayWithObject: path] withGI:gi toFile:outFile ofType:outputType withFlags:flags fileMode:fileMode refreshTime: &curTime];
                    if (NSApplicationDelegateReplySuccess != exitStatus)
                        break;
                }
            }
            
#if 0
            exists = [fm fileExistsAtPath: outFile isDirectory: &isFolder];
            if (exists && isFolder)
            {
                [pool release];
                continue;
            }
            
            if (exists)
            {
                [progressController showReplaceConfirmIfNecessaryForFile: [outFile lastPathComponent]];
                [progressController refresh];
                theReplaceChoice = [progressController replaceChoice];
                
                if (kSquishStop == theReplaceChoice)
                {
                    [pool release];
                    break;
                }
                else if (kSquishNoReplace == theReplaceChoice)
                {
                    [pool release];
                    continue;
                }
            }
            
            curTime = [[NSDate alloc] init];
            if ([curTime timeIntervalSinceDate: oldTime] > 1.0)
            {
                if (![progressController refresh])
                {
                    [pool release];
                    break;
                }
                [oldTime release];
                oldTime = curTime;
                curTime = nil;
            }
            
            [dc noteNewRecentDocumentURL: [NSURL fileURLWithPath: path]];
            
//        NSDate *startFileDate = [NSDate date];
            exitStatus = [self exportFile: path
                                   toFile: outFile
                                   ofType: outputType
                                withFlags: flags];
//        NSLog(@"total time for '%@' = %f s", outFile, [[NSDate date] timeIntervalSinceDate: startFileDate]);
#endif
            
            [pool release];
        }
    }

    [curTime release];

    [progressController endProgress];
    theReplaceChoice = [progressController replaceChoice];
    [progressController close];
    [progressController release];
    progressController = nil;

    if (NSApplicationDelegateReplyFailure == exitStatus)
    {
        NSRunInformationalAlertPanel(NSLocalizedString(@"Bad stuff happened",""), [NSString stringWithFormat: NSLocalizedString(@"An error occured while processing%@.",""), (path ? [NSString stringWithFormat:@" \"%@\"", path] : @"")], nil, nil, nil);
    }

    return exitStatus;
}

- (NSApplicationDelegateReply)exportTilesConfirmingReplace: (NSArray *)aTiles
                                                    withGI: (GIWrapper *)aGi
                                                    toFile: (NSString *)aOutFile
                                                    ofType: (NSString *)aType
                                                 withFlags: (int)aFlags
                                                  fileMode: (int)aFileMode
                                               refreshTime: (NSDate **)aTime
{
    BOOL exists, isFolder;
    NSDate *curTime = nil;
    int theReplaceChoice = kSquishNoReplace;
    NSApplicationDelegateReply exitStatus = NSApplicationDelegateReplySuccess;

    exists = [[NSFileManager defaultManager] fileExistsAtPath: aOutFile isDirectory: &isFolder];
    if (exists && isFolder)
    {
        return NSApplicationDelegateReplySuccess;
    }
    
    if (exists)
    {
        [progressController showReplaceConfirmIfNecessaryForFile: [aOutFile lastPathComponent]];
        [progressController refresh];
        theReplaceChoice = [progressController replaceChoice];
        
        switch (theReplaceChoice)
        {
            case kSquishStop:
                return NSApplicationDelegateReplyCancel;
            case kSquishNoReplace:
                return NSApplicationDelegateReplySuccess;
            default:
                break;
        }
    }
    
    curTime = [[NSDate alloc] init];
    if ([curTime timeIntervalSinceDate: *aTime] > 1.0)
    {
        if (![progressController refresh])
        {
            [curTime release];
            return exitStatus;
        }
        [*aTime release];
        *aTime = curTime;
    }

    [[NSDocumentController sharedDocumentController] noteNewRecentDocumentURL: [NSURL fileURLWithPath: [aTiles objectAtIndex: 0]]];

//        NSDate *startFileDate = [NSDate date];
    exitStatus = [self exportTiles: aTiles
                            withGI: aGi
                            toFile: aOutFile
                            ofType: aType
                         withFlags: aFlags
                          fileMode: aFileMode];
//        NSLog(@"total time for '%@' = %f s", outFile, [[NSDate date] timeIntervalSinceDate: startFileDate]);
    return exitStatus;
}

- (NSApplicationDelegateReply)exportTiles: (NSArray *)aTiles
                                   withGI: (GIWrapper *)aGi
                                   toFile: (NSString *)aOutFile
                                   ofType: (NSString *)aType
                                withFlags: (int)aFlags
                                 fileMode: (int)aFileMode
{
    NSApplicationDelegateReply result = NSApplicationDelegateReplyFailure;

    NSUserDefaults *defaults = [NSUserDefaults standardUserDefaults];
    BOOL genMipmapsSetting = [defaults boolForKey: @"genMipmaps"];
    int mipmapFormat = [defaults integerForKey: @"mipmapFormat"];
    int premulOption = [defaults integerForKey: @"premulAlpha"];

    if (aGi)
        [aGi retain];
    OSErr err = noErr;
    ComponentInstance gi = NULL;
    ComponentResult importResult;
    CGImageRef cgImage;
    CMProfileRef embedProf = NULL;
    Handle embedProfileH = NULL;
    NSString *embedProfName = nil;
    int profileChoice = kSquishProfileUseEmbedded;

#ifdef USE_DDS_PREVIEW
    [[ddsPreview window] makeKeyAndOrderFront: self];
#endif

    NSEnumerator *tileEnum = [aTiles objectEnumerator];
    NSString *aFile;
    FILE *ofp = NULL;
    BOOL isMipTile = NO;
    unsigned char *contextData = nil;

    while ((aFile = [tileEnum nextObject]))
    {
        if (nil == aGi)
            aGi = [[GIWrapper alloc] initWithPath: aFile];
        if (nil == aGi)
            break;
        gi = [aGi gi];

        // Get the embedded color profile if any
        // kGraphicsImporterDontUseColorMatching is required to prevent
        // GraphicsImportGetColorSyncProfile() from returning a default
        // profile all the time
        GraphicsImportSetFlags(gi, kGraphicsImporterDontUseColorMatching);
        if (noErr == GraphicsImportGetColorSyncProfile(gi, &embedProfileH))
        {
            if (embedProfileH)
            {
                CMProfileLocation embedProfLoc;
                embedProfLoc.locType = cmHandleBasedProfile;
                embedProfLoc.u.handleLoc.h = embedProfileH;
                
                if (noErr == CMOpenProfile(&embedProf, &embedProfLoc))
                {
                    embedProfName = (NSString *)CopyProfileDescriptionCFString(embedProf);
                    if (embedProfName)
                    {
                        //                            NSLog(@"Embedded profile = '%@'", embedProfName);
                        [embedProfName autorelease];
                    }
                }
                
                DisposeHandle(embedProfileH);
            }
        }
        if (embedProf)
        {
            [progressController showProfileConfirmIfNecessaryForFile: [aFile lastPathComponent] embeddedProfileName: embedProfName];
            [progressController refresh];
            profileChoice = [progressController profileChoice];
            switch (profileChoice)
            {
                case kSquishStop:
                    CMCloseProfile(embedProf);
                    return NSApplicationDelegateReplyCancel;
                case kSquishProfileOverride:
                    //                    NSLog(@"Override choice: %d", [progressController profileOverrideChoice]);
                    break;
                case kSquishProfileUseEmbedded:
                default:
                    break;
            }
        }
    
        GraphicsImportSetFlags(gi, 0);
        BOOL colorMatchFailed = NO;

        if (kSquishProfileOverride == profileChoice)
        {
            switch ([progressController profileOverrideChoice])
            {
                case 0:
                {
                    // Making source and destination profiles the same
                    // effectively bypasses color matching
                    GraphicsImportSetOverrideSourceColorSyncProfileRef(gi, sysprof);
                    GraphicsImportSetDestinationColorSyncProfileRef(gi, sysprof);
                    break;
                }
                case cmSRGBData:
                {
                    CMProfileRef srgbprof = CreateICCsRGBProfile();
                    if (srgbprof)
                    {
                        GraphicsImportSetOverrideSourceColorSyncProfileRef(gi, srgbprof);
                        GraphicsImportSetDestinationColorSyncProfileRef(gi, sysprof);
                        CMCloseProfile(srgbprof);
                    }
                    else
                    {
                        colorMatchFailed = YES;
                    }
                    break;
                }
                default:
                    colorMatchFailed = YES;
            }
        }
        else if (embedProf)
        {
            GraphicsImportSetDestinationColorSyncProfileRef(gi, embedProf);
        }

        if (embedProf)
            CMCloseProfile(embedProf);

        if (colorMatchFailed)
        {
            return result;
        }
//            else
//                CFShow(cspace);
    
        importResult = GraphicsImportCreateCGImage(gi, &cgImage, kGraphicsImportCreateCGImageUsingCurrentSettings);
        [aGi release]; aGi = nil;

        if (noErr == importResult)
        {
            ComponentInstance ge = NULL;
            Handle dstDataRef = NULL;
            OSType dstDataRefType;
            OSType dstType;
            
            if (kFileModeMipmap == aFileMode)
            {
                switch (mipmapFormat)
                {
                    case kMipmapTIFF:
                        dstType = kQTFileTypeTIFF;
                        break;
                    case kMipmapPNG:
                    default:
                        dstType = kQTFileTypePNG;
                        break;
                }
                
                err = QTNewDataReferenceFromFullPathCFString((CFStringRef)aOutFile, (QTPathStyle)kQTNativeDefaultPathStyle, 0, &dstDataRef, &dstDataRefType);
                if (noErr != err)
                {
                    CGImageRelease(cgImage);
                    return NSApplicationDelegateReplyFailure;
                }
                
                err = OpenADefaultComponent(GraphicsExporterComponentType, dstType, &ge);
                if (noErr != err)
                {
                    CGImageRelease(cgImage);
                    return NSApplicationDelegateReplyFailure;
                }
            }
            
#ifdef USE_DDS_PREVIEW
            NSImage *beforeImage = [[NSImage alloc] initWithContentsOfFile: aFile];
            [ddsPreview setBeforeImage: [beforeImage autorelease]];
#endif
            size_t width = CGImageGetWidth(cgImage);
            size_t height = CGImageGetHeight(cgImage);
            size_t bitsPerComponent = 8; //fp formats not supported (yet?)
            int samplesPerPixel = 4;
            size_t bytesPerRow = samplesPerPixel * width;
            int alpha = 1; // 1=ARGB, -1=RGBA
#ifdef USE_AUTO_PREMULTIPLY
            CGImageAlphaInfo alphaInfo = CGImageGetAlphaInfo(cgImage);
            BOOL isPremultiplied = YES;
            
            switch (alphaInfo)
            {
                case kCGImageAlphaLast:
                case kCGImageAlphaFirst:
                    isPremultiplied = NO;
                    break;
                default:
                    break;
            }
#endif
            // TODO? Respect color profile for grayscale images
            /*
             if (samplesPerPixel == 1)
             {
             if (cspace)
             {
             CGColorSpaceRelease(cspace);
             cspace = NULL;
             }
             
             if (kCGImageAlphaOnly != alphaInfo)
             {
             CMProfileRef grayref = CreateICCGenericGrayProfile();
             if (grayref)
             {
             cspace = CGColorSpaceCreateWithPlatformColorSpace(grayref);
             CMCloseProfile(grayref);
             }
             }
             }
             */

            // Silently not generate mipmaps if not power of two
            BOOL genMipmaps = (genMipmapsSetting &&
                               (width % 2 == 0) && (height % 2 == 0));

            if (isMipTile && !genMipmaps)
            {
                break;
            }

            if (!isMipTile)
            {
                contextData = (unsigned char *)malloc(bytesPerRow * height);
                ofp = (kFileModeDDS == aFileMode) ? fopen([aOutFile UTF8String], "wb") : NULL;
                if (ofp)
                {
                    result = [self writeDDSHeaderToFile: ofp
                                                 ofType: aType
                                                  width: width
                                                 height: height
                                              withFlags: aFlags
                                                mipmaps: genMipmaps];
                    
                    if (NSApplicationDelegateReplySuccess != result)
                    {
                        fclose(ofp);
                        free(contextData);
                        return result;
                    }
                }
            }

            // Reuse contextData for all miplevels
            // (Cannot reuse the CGBitmapContext object because 
            // changing the image dimensions cause image coords
            // to fall out of sync with the memory layout)
            memset(contextData, 0, bytesPerRow * height);
            CGContextRef bitmapContext = CGBitmapContextCreate(contextData, width, height, bitsPerComponent, bytesPerRow, cspace, kCGImageAlphaPremultipliedFirst);
            // Squish expects a "alpha last" RGBA format, but if we are to 
            // support vectorized premuliplication down to OS X 10.3,
            // "alpha first" ARGB is the only option.
            CGContextDrawImage(bitmapContext, CGRectMake(0, 0, width, height), cgImage);

            if (kFileModeMipmap == aFileMode)
            {
                GraphicsExportSetInputCGBitmapContext(ge, bitmapContext);
                err = GraphicsExportSetOutputDataReference(ge, dstDataRef, dstDataRefType);
                if (noErr == err)
                    err = GraphicsExportDoExport(ge, NULL);
                CloseComponent(ge);
                DisposeHandle(dstDataRef);
                if (noErr == err)
                    result = NSApplicationDelegateReplySuccess;
            }
            else
            {
                result = [self compressBytes: contextData
                                      toFile: ofp
                                       width: width
                                      height: height
                                 bytesPerRow: bytesPerRow
                             samplesPerPixel: samplesPerPixel
                                       alpha: alpha
                                 premultiply:
#ifdef USE_AUTO_PREMULTIPLY
                          (!isPremultiplied)
#else
                          premulOption
#endif
                                   withFlags: aFlags ];
                
                if (genMipmaps && [aTiles count] < 2)
                {
                    size_t mipw = width;
                    size_t miph = height;
                    while (mipw > 1 || miph > 1)
                    {
                        if (mipw > 1) { mipw >>= 1; bytesPerRow >>= 1; }
                        if (miph > 1) { miph >>= 1; }
//                          NSLog(@"Mipmap %ld x %ld", mipw, miph);
                        CGContextRelease(bitmapContext);
                        // Reuse contextData for all miplevels
                        // (Cannot reuse the CGBitmapContext object because 
                        // changing the image dimensions cause image coords
                        // to fall out of sync with the memory layout)
                        memset(contextData, 0, bytesPerRow * miph);
                        bitmapContext = CGBitmapContextCreate(contextData, mipw, miph, bitsPerComponent, bytesPerRow, cspace, kCGImageAlphaPremultipliedFirst);
                        CGContextDrawImage(bitmapContext, CGRectMake(0, 0, mipw, miph), cgImage);
                        result = [self compressBytes: contextData
                                              toFile: ofp
                                               width: mipw
                                              height: miph
                                         bytesPerRow: bytesPerRow
                                     samplesPerPixel: samplesPerPixel
                                               alpha: alpha
                                         premultiply:
#ifdef USE_AUTO_PREMULTIPLY
                                  (!isPremultiplied)
#else
                                  premulOption
#endif
                                           withFlags: aFlags ];
                    }
                }
            }
            CGContextRelease(bitmapContext);
            CGImageRelease(cgImage);
        }
        isMipTile = YES;    // subsequent files are mipmap tiles
    }

    if (ofp) fclose(ofp);
    free(contextData);
    return result;
}


- (NSApplicationDelegateReply)compressBytes: (unsigned char *)aBytes
                                    toBlock: (u8 *)output
                                      width: (size_t)aWidth
                                     height: (size_t)aHeight
                                bytesPerRow: (size_t)aBytesPerRow
                            samplesPerPixel: (int)aSamplesPerPixel
                                      alpha: (int)aAlpha
                                  withFlags: (int)aFlags
{
    NSApplicationDelegateReply result = NSApplicationDelegateReplyFailure;
    if (NULL == output) return result;
    SquishCompressNotifier progressNotifier(progressController);

    CompressImage(aBytes,
                  aWidth, aHeight,
                  aBytesPerRow, aSamplesPerPixel,
                  aAlpha,
                  output,
                  aFlags,
                  &progressNotifier);
    
    if (progressNotifier.isAborted())
    {
        return NSApplicationDelegateReplyCancel;
    }
    return NSApplicationDelegateReplySuccess;
}

- (NSApplicationDelegateReply)writeDDSHeaderToFile: (FILE *)ofp
                                            ofType: (NSString *)aType
                                             width: (size_t)aWidth
                                            height: (size_t)aHeight
                                         withFlags: (int)aFlags
                                           mipmaps: (BOOL)genMipmaps
{
    NSApplicationDelegateReply result = NSApplicationDelegateReplyFailure;
    if (NULL == ofp) return result;

    int byteSize = GetStorageRequirements(aWidth, aHeight, aFlags);
    uint32_t mipLevels = 1;
    uint32_t ddsFlags = DDSD_CAPS | DDSD_PIXELFORMAT | DDSD_WIDTH | DDSD_HEIGHT | DDSD_LINEARSIZE;
    uint32_t ddsCaps = DDSCAPS_TEXTURE;
    
    if (genMipmaps)
    {
        size_t mipw = aWidth;
        size_t miph = aHeight;
        while (mipw > 1 || miph > 1)
        {
            if (mipw > 1) mipw >>= 1;
            if (miph > 1) miph >>= 1;
            byteSize += GetStorageRequirements(mipw, miph, aFlags);
            ++mipLevels;
        }
        if (mipLevels > 1)
        {
            ddsFlags |= DDSD_MIPMAPCOUNT;
            ddsCaps  |= DDSCAPS_COMPLEX | DDSCAPS_MIPMAP;
        }
    }

    char ddsMagic[4];
    ddsMagic[0] = 'D';
    ddsMagic[1] = 'D';
    ddsMagic[2] = 'S';
    ddsMagic[3] = ' ';
    size_t ddsmsize = sizeof ddsMagic;
    fwrite(ddsMagic, ddsmsize, 1, ofp);
    
    DDSurfaceDesc ddsd = {0};
    ddsd.size = CFSwapInt32HostToLittle(sizeof ddsd);
    ddsd.flags = CFSwapInt32HostToLittle(ddsFlags);
    ddsd.width = CFSwapInt32HostToLittle(aWidth);
    ddsd.height = CFSwapInt32HostToLittle(aHeight);
    ddsd.pitch = CFSwapInt32HostToLittle(byteSize);
    ddsd.mipMapLevels = CFSwapInt32HostToLittle(mipLevels);
    ddsd.format.size = CFSwapInt32HostToLittle(sizeof ddsd.format);
    ddsd.format.flags = CFSwapInt32HostToLittle(DDPF_FOURCC);
    ddsd.format.fourCC = CFSwapInt32HostToLittle(FourCC([aType UTF8String]));
    ddsd.caps.caps = CFSwapInt32HostToLittle(ddsCaps);
    
    fwrite(&ddsd, sizeof ddsd, 1, ofp);

    return NSApplicationDelegateReplySuccess;
}

- (NSApplicationDelegateReply)compressBytes: (unsigned char *)aBytes
                                     toFile: (FILE *)ofp
                                      width: (size_t)aWidth
                                     height: (size_t)aHeight
                                bytesPerRow: (size_t)aBytesPerRow
                            samplesPerPixel: (int)aSamplesPerPixel
                                      alpha: (int)aAlpha
                                premultiply:
#ifdef USE_AUTO_PREMULTIPLY
                                             (BOOL)aPremultiply
#else
                                             (int)aPremulOption
#endif
                                  withFlags: (int)aFlags
{
    NSApplicationDelegateReply result = NSApplicationDelegateReplyFailure;
    int byteSize = GetStorageRequirements(aWidth, aHeight, aFlags);
    u8 *output = (u8 *)calloc(1, byteSize);
    if (NULL == output) return result;

    int format = [[NSUserDefaults standardUserDefaults] integerForKey: @"outputFormat"];
    if (kDxt5nm == format)
    {
#ifdef USE_AUTO_PREMULTIPLY
        aPremultiply = NO;
#else
        aPremulOption = kAlphaUnchanged;
#endif
        ArrangeNormalMapChannels(aBytes, aWidth, aHeight, aSamplesPerPixel);
    }

#ifdef USE_AUTO_PREMULTIPLY
    if (aPremultiply)
    {
        PremultiplyImageAlpha(aBytes, aWidth, aHeight, aBytesPerRow);
    }
#else
    switch (aPremulOption)
    {
        case kAlphaPremultiply:
            PremultiplyImageAlpha(aBytes, aWidth, aHeight, aBytesPerRow);
            break;
        case kAlphaUnpremultiply:
            UnpremultiplyImageAlpha(aBytes, aWidth, aHeight, aBytesPerRow);
            break;
        default:
            break;
    }
#endif

    result = [self compressBytes: aBytes
                         toBlock: output
                           width: aWidth
                          height: aHeight
                     bytesPerRow: aBytesPerRow
                 samplesPerPixel: aSamplesPerPixel
                           alpha: aAlpha
                       withFlags: aFlags];
    
    if (NSApplicationDelegateReplySuccess == result)
    {
        fwrite(output, byteSize, 1, ofp);
#ifdef USE_DDS_PREVIEW
        u8 *decomp = NULL;
        result = [self decompressBytes: output
                               toBlock: &decomp
                                 width: aWidth
                                height: aHeight
                             withFlags: aFlags];
        if (NSApplicationDelegateReplySuccess == result &&
            aWidth > 64)
        {
            NSBitmapImageRep *decompRep = [[NSBitmapImageRep alloc] initWithBitmapDataPlanes:nil pixelsWide:aWidth pixelsHigh:aHeight bitsPerSample:8 samplesPerPixel:aSamplesPerPixel hasAlpha:YES isPlanar:NO colorSpaceName:NSCalibratedRGBColorSpace bytesPerRow:aBytesPerRow bitsPerPixel:32];
            if (decompRep)
            {
                unsigned char *decompDestData = [decompRep bitmapData];
                memcpy(decompDestData, decomp, aBytesPerRow * aHeight);
                NSImage *decompImg = [[NSImage alloc] initWithSize: NSMakeSize(aWidth, aHeight)];
                if (decompImg)
                {
                    [decompImg addRepresentation: [decompRep autorelease]];
                    [decompImg setScalesWhenResized: YES];
                    [decompImg setSize: NSMakeSize(aWidth, aHeight)];
                    [ddsPreview setAfterImage: [decompImg autorelease]];
                }
            }
        }
        if (decomp)
            free(decomp);
#endif
    }

    free(output);
    return result;
}

- (NSApplicationDelegateReply)decompressBytes: (u8 *)aBytes
                                      toBlock: (u8 **)output
                                        width: (size_t)aWidth
                                       height: (size_t)aHeight
                                    withFlags: (int)aFlags
{
    NSApplicationDelegateReply result = NSApplicationDelegateReplyFailure;
    if (NULL == output) return result;
    *output = (u8 *)calloc(1, 4 * aWidth * aHeight);
    DecompressImage(*output, aWidth, aHeight, aBytes, aFlags);
    result = NSApplicationDelegateReplySuccess;
    return result;
}

- (NSArray *)extensions
{
    return [NSArray arrayWithObjects:
            @"bmp", @"gif", @"jpg", @"jpeg", @"jpe", @"png", @"tiff", @"tif",
            nil];
}

- (int)flagsFromPrefs
{
    NSUserDefaults *defaults = [NSUserDefaults standardUserDefaults];
    int flags = 0;
    int format = [defaults integerForKey: @"outputFormat"];
    flags  = (kDxt5nm == format) ? kDxt5 : format;
    flags |= [defaults integerForKey: @"quality"];
    if ([defaults boolForKey: @"weightByAlpha"])
        flags |= kWeightColourByAlpha;
    flags |= [defaults integerForKey: @"metric"];

    return flags;
}

- (NSString *)pathForOutputFile: (NSString *)aFile
                         suffix: (NSString *)aSuffix
                      extension: (NSString *)aExt
                   outputFolder: (NSString *)aDest
{
    NSString *result = [[[aFile stringByDeletingPathExtension] stringByAppendingString:aSuffix] stringByAppendingPathExtension: aExt];
    if (aDest)
    {
        result = [aDest stringByAppendingPathComponent: [result lastPathComponent]];
    }
    return result;
}


- (void)browseDestination: (id)aSender
{
    NSUserDefaults *defaults = [NSUserDefaults standardUserDefaults];
    NSString *selection = nil;
    NSOpenPanel *oPanel = [NSOpenPanel openPanel];
    [oPanel setAllowsMultipleSelection: NO];
    [oPanel setCanChooseDirectories: YES];
    [oPanel setCanChooseFiles: NO];
    [oPanel setCanCreateDirectories: YES];

    if (NSFileHandlingPanelOKButton ==
        [oPanel runModalForDirectory:nil file:nil])
    {
        NSArray *filenames = [oPanel filenames];
        if (filenames && [filenames count] > 0)
            selection = [filenames objectAtIndex: 0];
    }

    [defaults setObject: selection forKey: @"browseDest"];
}

- (void)convert: (id)aSender
{
    [NSApp stopModal];
}

- (void)abortConvert: (id)aSender
{
    [NSApp abortModal];
}

- (void)showMyHelp: (id)aSender
{
    [helpWindow setWorksWhenModal: YES];
    [helpWindow makeKeyAndOrderFront: aSender];
}

- (BOOL)enableGenMipmaps
{
    return enableGenMipmaps;
}

- (BOOL)enableAssembleMipmaps
{
    return enableAssembleMipmaps;
}
- (void)setEnableGenMipmaps: (BOOL)aEnable { enableGenMipmaps = aEnable; }
- (void)setEnableAssembleMipmaps: (BOOL)aEnable { enableAssembleMipmaps = aEnable; }
@end
