//
//  SquishDocument.m
//  squish
//
//  Created by Da Woon Jung on 6/22/06.
//

#import "SquishDocument.h"
#import <squish/squish.h>
#import "dds.h"
#import "SquishCompressNotifier.h"

using namespace squish;


@interface SquishDocument (Private)
- (int)setFlagsFromPrefs: (int)aFlags;
@end


@implementation SquishDocument

- (id)init
{
    self = [super initWithWindowNibName: @"SquishDocument"];
    if (self)
    {
    }

    return self;
}

- (void)dealloc
{
    [super dealloc];
}


- (BOOL)writeToFile:(NSString *)aFileName ofType:(NSString *)aType
{
    BOOL result = NO;

    if (aType)
    {
        int flags = kDxt1;

        if ([aType isEqualToString: @"DXT1"])
        {
//            NSLog(@"DXT1");
            flags = kDxt1;
        }
        else if ([aType isEqualToString: @"DXT3"])
        {
//            NSLog(@"DXT3");
            flags = kDxt3;
        }
        else if ([aType isEqualToString: @"DXT5"])
        {
//            NSLog(@"DXT5");
            flags = kDxt5;
        }

        flags = [self setFlagsFromPrefs: flags];

        NSArray *reps = [image representations];
        if (reps && [reps count] > 0)
        {
            NSEnumerator *repEnum = [reps objectEnumerator];
            NSImageRep *rep = nil;
            NSBitmapImageRep *bitmap = nil;
            unsigned char *bitmapData = nil;
            int width = 0, height = 0;
            int byteSize = 0;
            u8 *output = nil;

            while ((rep = [repEnum nextObject]))
            {
                if ([rep isKindOfClass: [NSBitmapImageRep class]])
                {
                    NSWindow *ourWindow = [imageView window];
                    NSModalSession progressSession = [NSApp beginModalSessionForWindow: ourWindow];
                    SquishCompressNotifier progressNotifier(progressBar, progressSession);
                    
                    [NSApp beginSheet: progressPanel
                       modalForWindow: [imageView window]
                        modalDelegate: self
                       didEndSelector: nil contextInfo: nil];

                    [progressBar setIndeterminate: NO];

                    bitmap = (NSBitmapImageRep *)rep;
//                    NSLog(@"Planar? %d (%d bytes/plane, %d bytes/row, %d samplesPerPixel)", [bitmap isPlanar], [bitmap bytesPerPlane], [bitmap bytesPerRow], [bitmap samplesPerPixel]);
                    // Only meshed format is supported for now
                    bitmapData = [bitmap bitmapData]; //(unsigned char *)calloc(1, [bitmap bytesPerPlane]);
//                    memcpy(bitmapData, [bitmap bitmapData], [bitmap bytesPerPlane]);

                    width = [bitmap pixelsWide];
                    height = [bitmap pixelsHigh];
                    byteSize = GetStorageRequirements(width, height, flags);
                    output = (u8 *)calloc(1, byteSize);

                    CompressImage(bitmapData,
                                  width,
                                  height,
                                  [bitmap bytesPerRow],
                                  [bitmap samplesPerPixel],
                                  output,
                                  flags | kWeightColourByAlpha,
                                  &progressNotifier);

                    [progressBar setIndeterminate: YES];
                    [progressBar startAnimation: self];

                    FILE *ofp = fopen([aFileName UTF8String], "wb");
                    if (ofp)
                    {
                        // write the header
                        char ddsMagic[4];
                        ddsMagic[0] = 'D';
                        ddsMagic[1] = 'D';
                        ddsMagic[2] = 'S';
                        ddsMagic[3] = ' ';
                        size_t ddsmsize = sizeof ddsMagic;
                        fwrite(ddsMagic, ddsmsize, 1, ofp);

                        DDSurfaceDesc ddsd = {0};
                        ddsd.size = CFSwapInt32HostToLittle(sizeof ddsd);
                        ddsd.flags = CFSwapInt32HostToLittle(DDSD_CAPS | DDSD_PIXELFORMAT | DDSD_WIDTH | DDSD_HEIGHT | DDSD_LINEARSIZE);
                        ddsd.width = CFSwapInt32HostToLittle(width);
                        ddsd.height = CFSwapInt32HostToLittle(height);
                        ddsd.pitch = CFSwapInt32HostToLittle(byteSize);
//                        ddsd.mipMapLevels = CFSwapInt32HostToLittle(1U);
                        ddsd.format.size = CFSwapInt32HostToLittle(sizeof ddsd.format);
                        ddsd.format.flags = CFSwapInt32HostToLittle(DDPF_FOURCC);
                        ddsd.format.fourCC = CFSwapInt32HostToLittle(FourCC([aType UTF8String]));
//                        ddsd.format.bpp = CFSwapInt32HostToLittle([bitmap bitsPerPixel]);
                        ddsd.caps.caps = CFSwapInt32HostToLittle(DDSCAPS_TEXTURE);

                        fwrite(&ddsd, sizeof ddsd, 1, ofp);

                        // write the data
                        fwrite(output, byteSize, 1, ofp);

                        fclose(ofp);
                        result = YES;
                    }

                    free(output);
//                    free(bitmapData);

                    [progressBar stopAnimation: self];
                    [NSApp endModalSession: progressSession];
                    [progressPanel orderOut: self];
                    break;
                }
            }
        }
    }
    
    return result;
}

- (BOOL)readFromFile:(NSString *)aFilename ofType:(NSString *)aDocType
{
    BOOL result = NO;

    NSImage *img = [[NSImage alloc] initWithContentsOfFile:aFilename];
    if (img)
    {
        result = YES;
        image = img;
    }

    return result;
}

- (int)setFlagsFromPrefs: (int)aFlags
{
    NSUserDefaults *defaults = [NSUserDefaults standardUserDefaults];
    int flags = aFlags;

    flags |= [defaults integerForKey: @"quality"];
    flags |= [defaults integerForKey: @"weightByAlpha"];
    flags |= [defaults integerForKey: @"metric"];

    return flags;
}

@end
